@@ -2,6 +2,8 @@ package registry
 
 import (
 	"os"
+	"path/filepath"
+	"strings"
 	"sync"
 
 	"github.com/containers/podman/v4/pkg/domain/entities"
@@ -15,9 +17,18 @@ var remoteFromCLI = struct {
 	sync  sync.Once
 }{}
 
+const PodmanSh = "podmansh"
+
 // IsRemote returns true if podman was built to run remote or --remote flag given on CLI
 // Use in init() functions as an initialization check
 func IsRemote() bool {
+	// remote conflicts with podmansh in how the `-c` option gets parsed
+	// This is noticeable if a user with shell set to podmansh were to execute
+	// a command using ssh like so:
+	// ssh user@host id
+	if strings.HasSuffix(filepath.Base(os.Args[0]), PodmanSh) {
+		return false
+	}
 	remoteFromCLI.sync.Do(func() {
 		remote := false
 		if _, ok := os.LookupEnv("CONTAINER_HOST"); ok {
@@ -30,6 +41,14 @@ func IsRemote() bool {
 		fs.Usage = func() {}
 		fs.SetInterspersed(false)
 		fs.BoolVarP(&remoteFromCLI.Value, "remote", "r", remote, "")
+		connectionFlagName := "connection"
+		fs.StringP(connectionFlagName, "c", "", "")
+		contextFlagName := "context"
+		fs.String(contextFlagName, "", "")
+		hostFlagName := "host"
+		fs.StringP(hostFlagName, "H", "", "")
+		urlFlagName := "url"
+		fs.String(urlFlagName, "", "")
 
 		// The shell completion logic will call a command called "__complete" or "__completeNoDesc"
 		// This command will always be the second argument
@@ -39,6 +58,8 @@ func IsRemote() bool {
 			start = 2
 		}
 		_ = fs.Parse(os.Args[start:])
+		// --connection or --url implies --remote
+		remoteFromCLI.Value = remoteFromCLI.Value || fs.Changed(connectionFlagName) || fs.Changed(urlFlagName) || fs.Changed(hostFlagName) || fs.Changed(contextFlagName)
 	})
 	return podmanOptions.EngineMode == entities.TunnelMode || remoteFromCLI.Value
 }
