@@ -132,6 +132,8 @@ func isExtSupported(filename string) bool {
 	return ok
 }
 
+var seen = make(map[string]bool)
+
 func loadUnitsFromDir(sourcePath string, units map[string]*parser.UnitFile) error {
 	var prevError error
 	files, err := os.ReadDir(sourcePath)
@@ -144,6 +146,9 @@ func loadUnitsFromDir(sourcePath string, units map[string]*parser.UnitFile) erro
 
 	for _, file := range files {
 		name := file.Name()
+		if seen[name] {
+			continue
+		}
 		if units[name] == nil && isExtSupported(name) {
 			path := path.Join(sourcePath, name)
 
@@ -156,6 +161,7 @@ func loadUnitsFromDir(sourcePath string, units map[string]*parser.UnitFile) erro
 				}
 			} else {
 				units[name] = f
+				seen[name] = true
 			}
 		}
 	}
