@@ -1,23 +1,27 @@
 package libpod
 
 import (
-	"io/ioutil"
+	"context"
+	"errors"
+	"fmt"
 	"net/http"
 	"os"
 	"path/filepath"
 	"strconv"
 	"time"
 
-	"github.com/containers/podman/v3/libpod/define"
-	"github.com/containers/podman/v3/libpod/events"
+	"github.com/containers/common/pkg/resize"
+	"github.com/containers/common/pkg/util"
+	"github.com/containers/podman/v4/libpod/define"
+	"github.com/containers/podman/v4/libpod/events"
 	"github.com/containers/storage/pkg/stringid"
-	"github.com/pkg/errors"
 	"github.com/sirupsen/logrus"
+	"golang.org/x/sys/unix"
 )
 
 // ExecConfig contains the configuration of an exec session
 type ExecConfig struct {
-	// Command the the command that will be invoked in the exec session.
+	// Command is the command that will be invoked in the exec session.
 	// Must not be empty.
 	Command []string `json:"command"`
 	// Terminal is whether the exec session will allocate a pseudoterminal.
@@ -77,11 +81,11 @@ type ExecConfig struct {
 type ExecSession struct {
 	// Id is the ID of the exec session.
 	// Named somewhat strangely to not conflict with ID().
-	// nolint:stylecheck,golint
+	//nolint:stylecheck,revive
 	Id string `json:"id"`
 	// ContainerId is the ID of the container this exec session belongs to.
 	// Named somewhat strangely to not conflict with ContainerID().
-	// nolint:stylecheck,golint
+	//nolint:stylecheck,revive
 	ContainerId string `json:"containerId"`
 
 	// State is the state of the exec session.
@@ -110,7 +114,7 @@ func (e *ExecSession) ContainerID() string {
 // configuration and current state.
 func (e *ExecSession) Inspect() (*define.InspectExecSession, error) {
 	if e.Config == nil {
-		return nil, errors.Wrapf(define.ErrInternal, "given exec session does not have a configuration block")
+		return nil, fmt.Errorf("given exec session does not have a configuration block: %w", define.ErrInternal)
 	}
 
 	output := new(define.InspectExecSession)
@@ -163,22 +167,22 @@ func (c *Container) ExecCreate(config *ExecConfig) (string, error) {
 
 	// Verify our config
 	if config == nil {
-		return "", errors.Wrapf(define.ErrInvalidArg, "must provide a configuration to ExecCreate")
+		return "", fmt.Errorf("must provide a configuration to ExecCreate: %w", define.ErrInvalidArg)
 	}
 	if len(config.Command) == 0 {
-		return "", errors.Wrapf(define.ErrInvalidArg, "must provide a non-empty command to start an exec session")
+		return "", fmt.Errorf("must provide a non-empty command to start an exec session: %w", define.ErrInvalidArg)
 	}
 	if config.ExitCommandDelay > 0 && len(config.ExitCommand) == 0 {
-		return "", errors.Wrapf(define.ErrInvalidArg, "must provide a non-empty exit command if giving an exit command delay")
+		return "", fmt.Errorf("must provide a non-empty exit command if giving an exit command delay: %w", define.ErrInvalidArg)
 	}
 
 	// Verify that we are in a good state to continue
 	if !c.ensureState(define.ContainerStateRunning) {
-		return "", errors.Wrapf(define.ErrCtrStateInvalid, "can only create exec sessions on running containers")
+		return "", fmt.Errorf("can only create exec sessions on running containers: %w", define.ErrCtrStateInvalid)
 	}
 
 	// Generate an ID for our new exec session
-	sessionID := stringid.GenerateNonCryptoID()
+	sessionID := stringid.GenerateRandomID()
 	found := true
 	// This really ought to be a do-while, but Go doesn't have those...
 	for found {
@@ -190,7 +194,7 @@ func (c *Container) ExecCreate(config *ExecConfig) (string, error) {
 			}
 		}
 		if found {
-			sessionID = stringid.GenerateNonCryptoID()
+			sessionID = stringid.GenerateRandomID()
 		}
 	}
 
@@ -201,7 +205,7 @@ func (c *Container) ExecCreate(config *ExecConfig) (string, error) {
 	session.State = define.ExecStateCreated
 	session.Config = new(ExecConfig)
 	if err := JSONDeepCopy(config, session.Config); err != nil {
-		return "", errors.Wrapf(err, "error copying exec configuration into exec session")
+		return "", fmt.Errorf("copying exec configuration into exec session: %w", err)
 	}
 
 	if len(session.Config.ExitCommand) > 0 {
@@ -241,16 +245,16 @@ func (c *Container) ExecStart(sessionID string) error {
 
 	// Verify that we are in a good state to continue
 	if !c.ensureState(define.ContainerStateRunning) {
-		return errors.Wrapf(define.ErrCtrStateInvalid, "can only start exec sessions when their container is running")
+		return fmt.Errorf("can only start exec sessions when their container is running: %w", define.ErrCtrStateInvalid)
 	}
 
 	session, ok := c.state.ExecSessions[sessionID]
 	if !ok {
-		return errors.Wrapf(define.ErrNoSuchExecSession, "container %s has no exec session with ID %s", c.ID(), sessionID)
+		return fmt.Errorf("container %s has no exec session with ID %s: %w", c.ID(), sessionID, define.ErrNoSuchExecSession)
 	}
 
 	if session.State != define.ExecStateCreated {
-		return errors.Wrapf(define.ErrExecSessionStateInvalid, "can only start created exec sessions, while container %s session %s state is %q", c.ID(), session.ID(), session.State.String())
+		return fmt.Errorf("can only start created exec sessions, while container %s session %s state is %q: %w", c.ID(), session.ID(), session.State.String(), define.ErrExecSessionStateInvalid)
 	}
 
 	logrus.Infof("Going to start container %s exec session %s and attach to it", c.ID(), session.ID())
@@ -275,10 +279,13 @@ func (c *Container) ExecStart(sessionID string) error {
 	return c.save()
 }
 
+func (c *Container) ExecStartAndAttach(sessionID string, streams *define.AttachStreams, newSize *resize.TerminalSize) error {
+	return c.execStartAndAttach(sessionID, streams, newSize, false)
+}
+
 // ExecStartAndAttach starts and attaches to an exec session in a container.
-// TODO: Should we include detach keys in the signature to allow override?
-// TODO: How do we handle AttachStdin/AttachStdout/AttachStderr?
-func (c *Container) ExecStartAndAttach(sessionID string, streams *define.AttachStreams) error {
+// newSize resizes the tty to this size before the process is started, must be nil if the exec session has no tty
+func (c *Container) execStartAndAttach(sessionID string, streams *define.AttachStreams, newSize *resize.TerminalSize, isHealthcheck bool) error {
 	if !c.batched {
 		c.lock.Lock()
 		defer c.lock.Unlock()
@@ -290,16 +297,16 @@ func (c *Container) ExecStartAndAttach(sessionID string, streams *define.AttachS
 
 	// Verify that we are in a good state to continue
 	if !c.ensureState(define.ContainerStateRunning) {
-		return errors.Wrapf(define.ErrCtrStateInvalid, "can only start exec sessions when their container is running")
+		return fmt.Errorf("can only start exec sessions when their container is running: %w", define.ErrCtrStateInvalid)
 	}
 
 	session, ok := c.state.ExecSessions[sessionID]
 	if !ok {
-		return errors.Wrapf(define.ErrNoSuchExecSession, "container %s has no exec session with ID %s", c.ID(), sessionID)
+		return fmt.Errorf("container %s has no exec session with ID %s: %w", c.ID(), sessionID, define.ErrNoSuchExecSession)
 	}
 
 	if session.State != define.ExecStateCreated {
-		return errors.Wrapf(define.ErrExecSessionStateInvalid, "can only start created exec sessions, while container %s session %s state is %q", c.ID(), session.ID(), session.State.String())
+		return fmt.Errorf("can only start created exec sessions, while container %s session %s state is %q: %w", c.ID(), session.ID(), session.State.String(), define.ErrExecSessionStateInvalid)
 	}
 
 	logrus.Infof("Going to start container %s exec session %s and attach to it", c.ID(), session.ID())
@@ -309,12 +316,17 @@ func (c *Container) ExecStartAndAttach(sessionID string, streams *define.AttachS
 		return err
 	}
 
-	pid, attachChan, err := c.ociRuntime.ExecContainer(c, session.ID(), opts, streams)
+	pid, attachChan, err := c.ociRuntime.ExecContainer(c, session.ID(), opts, streams, newSize)
 	if err != nil {
 		return err
 	}
 
-	c.newContainerEvent(events.Exec)
+	if isHealthcheck {
+		c.newContainerEvent(events.HealthStatus)
+	} else {
+		c.newContainerEvent(events.Exec)
+	}
+
 	logrus.Debugf("Successfully started exec session %s in container %s", session.ID(), c.ID())
 
 	var lastErr error
@@ -338,22 +350,60 @@ func (c *Container) ExecStartAndAttach(sessionID string, streams *define.AttachS
 	}
 	lastErr = tmpErr
 
-	exitCode, err := c.readExecExitCode(session.ID())
-	if err != nil {
+	exitCode, exitCodeErr := c.readExecExitCode(session.ID())
+
+	// Lock again.
+	// Important: we must lock and sync *before* the above error is handled.
+	// We need info from the database to handle the error.
+	if !c.batched {
+		c.lock.Lock()
+	}
+	// We can't reuse the old exec session (things may have changed from
+	// other use, the container was unlocked).
+	// So re-sync and get a fresh copy.
+	// If we can't do this, no point in continuing, any attempt to save
+	// would write garbage to the DB.
+	if err := c.syncContainer(); err != nil {
+		if errors.Is(err, define.ErrNoSuchCtr) || errors.Is(err, define.ErrCtrRemoved) {
+			// We can't save status, but since the container has
+			// been entirely removed, we don't have to; exit cleanly
+			return lastErr
+		}
 		if lastErr != nil {
 			logrus.Errorf("Container %s exec session %s error: %v", c.ID(), session.ID(), lastErr)
 		}
-		lastErr = err
+		return fmt.Errorf("syncing container %s state to update exec session %s: %w", c.ID(), sessionID, err)
+	}
+
+	// Now handle the error from readExecExitCode above.
+	if exitCodeErr != nil {
+		newSess, ok := c.state.ExecSessions[sessionID]
+		if !ok {
+			// The exec session was removed entirely, probably by
+			// the cleanup process. When it did so, it should have
+			// written an event with the exit code.
+			// Given that, there's nothing more we can do.
+			logrus.Infof("Container %s exec session %s already removed", c.ID(), session.ID())
+			return lastErr
+		}
+
+		if newSess.State == define.ExecStateStopped {
+			// Exec session already cleaned up.
+			// Exit code should be recorded, so it's OK if we were
+			// not able to read it.
+			logrus.Infof("Container %s exec session %s already cleaned up", c.ID(), session.ID())
+			return lastErr
+		}
+
+		if lastErr != nil {
+			logrus.Errorf("Container %s exec session %s error: %v", c.ID(), session.ID(), lastErr)
+		}
+		lastErr = exitCodeErr
 	}
 
 	logrus.Debugf("Container %s exec session %s completed with exit code %d", c.ID(), session.ID(), exitCode)
 
-	// Lock again
-	if !c.batched {
-		c.lock.Lock()
-	}
-
-	if err := writeExecExitCode(c, session.ID(), exitCode); err != nil {
+	if err := justWriteExecExitCode(c, session.ID(), exitCode); err != nil {
 		if lastErr != nil {
 			logrus.Errorf("Container %s exec session %s error: %v", c.ID(), session.ID(), lastErr)
 		}
@@ -372,7 +422,9 @@ func (c *Container) ExecStartAndAttach(sessionID string, streams *define.AttachS
 }
 
 // ExecHTTPStartAndAttach starts and performs an HTTP attach to an exec session.
-func (c *Container) ExecHTTPStartAndAttach(sessionID string, r *http.Request, w http.ResponseWriter, streams *HTTPAttachStreams, detachKeys *string, cancel <-chan bool, hijackDone chan<- bool) error {
+// newSize resizes the tty to this size before the process is started, must be nil if the exec session has no tty
+func (c *Container) ExecHTTPStartAndAttach(sessionID string, r *http.Request, w http.ResponseWriter,
+	streams *HTTPAttachStreams, detachKeys *string, cancel <-chan bool, hijackDone chan<- bool, newSize *resize.TerminalSize) error {
 	// TODO: How do we combine streams with the default streams set in the exec session?
 
 	// Ensure that we don't leak a goroutine here
@@ -391,16 +443,16 @@ func (c *Container) ExecHTTPStartAndAttach(sessionID string, r *http.Request, w
 
 	session, ok := c.state.ExecSessions[sessionID]
 	if !ok {
-		return errors.Wrapf(define.ErrNoSuchExecSession, "container %s has no exec session with ID %s", c.ID(), sessionID)
+		return fmt.Errorf("container %s has no exec session with ID %s: %w", c.ID(), sessionID, define.ErrNoSuchExecSession)
 	}
 
 	// Verify that we are in a good state to continue
 	if !c.ensureState(define.ContainerStateRunning) {
-		return errors.Wrapf(define.ErrCtrStateInvalid, "can only start exec sessions when their container is running")
+		return fmt.Errorf("can only start exec sessions when their container is running: %w", define.ErrCtrStateInvalid)
 	}
 
 	if session.State != define.ExecStateCreated {
-		return errors.Wrapf(define.ErrExecSessionStateInvalid, "can only start created exec sessions, while container %s session %s state is %q", c.ID(), session.ID(), session.State.String())
+		return fmt.Errorf("can only start created exec sessions, while container %s session %s state is %q: %w", c.ID(), session.ID(), session.State.String(), define.ErrExecSessionStateInvalid)
 	}
 
 	logrus.Infof("Going to start container %s exec session %s and attach to it", c.ID(), session.ID())
@@ -411,7 +463,7 @@ func (c *Container) ExecHTTPStartAndAttach(sessionID string, r *http.Request, w
 		session.ExitCode = define.ExecErrorCodeGeneric
 
 		if err := c.save(); err != nil {
-			logrus.Errorf("Error saving container %s exec session %s after failure to prepare: %v", err, c.ID(), session.ID())
+			logrus.Errorf("Saving container %s exec session %s after failure to prepare: %v", err, c.ID(), session.ID())
 		}
 
 		return err
@@ -430,13 +482,13 @@ func (c *Container) ExecHTTPStartAndAttach(sessionID string, r *http.Request, w
 		close(holdConnOpen)
 	}()
 
-	pid, attachChan, err := c.ociRuntime.ExecContainerHTTP(c, session.ID(), execOpts, r, w, streams, cancel, hijackDone, holdConnOpen)
+	pid, attachChan, err := c.ociRuntime.ExecContainerHTTP(c, session.ID(), execOpts, r, w, streams, cancel, hijackDone, holdConnOpen, newSize)
 	if err != nil {
 		session.State = define.ExecStateStopped
 		session.ExitCode = define.TranslateExecErrorToExitCode(define.ExecErrorCodeGeneric, err)
 
 		if err := c.save(); err != nil {
-			logrus.Errorf("Error saving container %s exec session %s after failure to start: %v", err, c.ID(), session.ID())
+			logrus.Errorf("Saving container %s exec session %s after failure to start: %v", err, c.ID(), session.ID())
 		}
 
 		return err
@@ -517,11 +569,11 @@ func (c *Container) ExecStop(sessionID string, timeout *uint) error {
 
 	session, ok := c.state.ExecSessions[sessionID]
 	if !ok {
-		return errors.Wrapf(define.ErrNoSuchExecSession, "container %s has no exec session with ID %s", c.ID(), sessionID)
+		return fmt.Errorf("container %s has no exec session with ID %s: %w", c.ID(), sessionID, define.ErrNoSuchExecSession)
 	}
 
 	if session.State != define.ExecStateRunning {
-		return errors.Wrapf(define.ErrExecSessionStateInvalid, "container %s exec session %s is %q, can only stop running sessions", c.ID(), session.ID(), session.State.String())
+		return fmt.Errorf("container %s exec session %s is %q, can only stop running sessions: %w", c.ID(), session.ID(), session.State.String(), define.ErrExecSessionStateInvalid)
 	}
 
 	logrus.Infof("Stopping container %s exec session %s", c.ID(), session.ID())
@@ -539,24 +591,13 @@ func (c *Container) ExecStop(sessionID string, timeout *uint) error {
 	var cleanupErr error
 
 	// Retrieve exit code and update status
-	exitCode, err := c.readExecExitCode(session.ID())
-	if err != nil {
-		cleanupErr = err
-	}
-	session.ExitCode = exitCode
-	session.PID = 0
-	session.State = define.ExecStateStopped
-
-	if err := c.save(); err != nil {
-		if cleanupErr != nil {
-			logrus.Errorf("Error stopping container %s exec session %s: %v", c.ID(), session.ID(), cleanupErr)
-		}
+	if err := retrieveAndWriteExecExitCode(c, session.ID()); err != nil {
 		cleanupErr = err
 	}
 
 	if err := c.cleanupExecBundle(session.ID()); err != nil {
 		if cleanupErr != nil {
-			logrus.Errorf("Error stopping container %s exec session %s: %v", c.ID(), session.ID(), cleanupErr)
+			logrus.Errorf("Stopping container %s exec session %s: %v", c.ID(), session.ID(), cleanupErr)
 		}
 		cleanupErr = err
 	}
@@ -578,7 +619,7 @@ func (c *Container) ExecCleanup(sessionID string) error {
 
 	session, ok := c.state.ExecSessions[sessionID]
 	if !ok {
-		return errors.Wrapf(define.ErrNoSuchExecSession, "container %s has no exec session with ID %s", c.ID(), sessionID)
+		return fmt.Errorf("container %s has no exec session with ID %s: %w", c.ID(), sessionID, define.ErrNoSuchExecSession)
 	}
 
 	if session.State == define.ExecStateRunning {
@@ -589,18 +630,10 @@ func (c *Container) ExecCleanup(sessionID string) error {
 		}
 
 		if alive {
-			return errors.Wrapf(define.ErrExecSessionStateInvalid, "cannot clean up container %s exec session %s as it is running", c.ID(), session.ID())
+			return fmt.Errorf("cannot clean up container %s exec session %s as it is running: %w", c.ID(), session.ID(), define.ErrExecSessionStateInvalid)
 		}
 
-		exitCode, err := c.readExecExitCode(session.ID())
-		if err != nil {
-			return err
-		}
-		session.ExitCode = exitCode
-		session.PID = 0
-		session.State = define.ExecStateStopped
-
-		if err := c.save(); err != nil {
+		if err := retrieveAndWriteExecExitCode(c, session.ID()); err != nil {
 			return err
 		}
 	}
@@ -624,7 +657,7 @@ func (c *Container) ExecRemove(sessionID string, force bool) error {
 
 	session, ok := c.state.ExecSessions[sessionID]
 	if !ok {
-		return errors.Wrapf(define.ErrNoSuchExecSession, "container %s has no exec session with ID %s", c.ID(), sessionID)
+		return fmt.Errorf("container %s has no exec session with ID %s: %w", c.ID(), sessionID, define.ErrNoSuchExecSession)
 	}
 
 	logrus.Infof("Removing container %s exec session %s", c.ID(), session.ID())
@@ -637,15 +670,15 @@ func (c *Container) ExecRemove(sessionID string, force bool) error {
 			return err
 		}
 		if !running {
-			session.State = define.ExecStateStopped
-			// TODO: should we retrieve exit code here?
-			// TODO: Might be worth saving state here.
+			if err := retrieveAndWriteExecExitCode(c, session.ID()); err != nil {
+				return err
+			}
 		}
 	}
 
 	if session.State == define.ExecStateRunning {
 		if !force {
-			return errors.Wrapf(define.ErrExecSessionStateInvalid, "container %s exec session %s is still running, cannot remove", c.ID(), session.ID())
+			return fmt.Errorf("container %s exec session %s is still running, cannot remove: %w", c.ID(), session.ID(), define.ErrExecSessionStateInvalid)
 		}
 
 		// Stop the session
@@ -653,6 +686,10 @@ func (c *Container) ExecRemove(sessionID string, force bool) error {
 			return err
 		}
 
+		if err := retrieveAndWriteExecExitCode(c, session.ID()); err != nil {
+			return err
+		}
+
 		if err := c.cleanupExecBundle(session.ID()); err != nil {
 			return err
 		}
@@ -675,7 +712,7 @@ func (c *Container) ExecRemove(sessionID string, force bool) error {
 
 // ExecResize resizes the TTY of the given exec session. Only available if the
 // exec session created a TTY.
-func (c *Container) ExecResize(sessionID string, newSize define.TerminalSize) error {
+func (c *Container) ExecResize(sessionID string, newSize resize.TerminalSize) error {
 	if !c.batched {
 		c.lock.Lock()
 		defer c.lock.Unlock()
@@ -687,13 +724,13 @@ func (c *Container) ExecResize(sessionID string, newSize define.TerminalSize) er
 
 	session, ok := c.state.ExecSessions[sessionID]
 	if !ok {
-		return errors.Wrapf(define.ErrNoSuchExecSession, "container %s has no exec session with ID %s", c.ID(), sessionID)
+		return fmt.Errorf("container %s has no exec session with ID %s: %w", c.ID(), sessionID, define.ErrNoSuchExecSession)
 	}
 
 	logrus.Infof("Resizing container %s exec session %s to %+v", c.ID(), session.ID(), newSize)
 
 	if session.State != define.ExecStateRunning {
-		return errors.Wrapf(define.ErrExecSessionStateInvalid, "cannot resize container %s exec session %s as it is not running", c.ID(), session.ID())
+		return fmt.Errorf("cannot resize container %s exec session %s as it is not running: %w", c.ID(), session.ID(), define.ErrExecSessionStateInvalid)
 	}
 
 	// The exec session may have exited since we last updated.
@@ -706,10 +743,10 @@ func (c *Container) ExecResize(sessionID string, newSize define.TerminalSize) er
 		session.State = define.ExecStateStopped
 
 		if err := c.save(); err != nil {
-			logrus.Errorf("Error saving state of container %s: %v", c.ID(), err)
+			logrus.Errorf("Saving state of container %s: %v", c.ID(), err)
 		}
 
-		return errors.Wrapf(define.ErrExecSessionStateInvalid, "cannot resize container %s exec session %s as it has stopped", c.ID(), session.ID())
+		return fmt.Errorf("cannot resize container %s exec session %s as it has stopped: %w", c.ID(), session.ID(), define.ErrExecSessionStateInvalid)
 	}
 
 	// Make sure the exec session is still running.
@@ -717,10 +754,14 @@ func (c *Container) ExecResize(sessionID string, newSize define.TerminalSize) er
 	return c.ociRuntime.ExecAttachResize(c, sessionID, newSize)
 }
 
+func (c *Container) Exec(config *ExecConfig, streams *define.AttachStreams, resize <-chan resize.TerminalSize) (int, error) {
+	return c.exec(config, streams, resize, false)
+}
+
 // Exec emulates the old Libpod exec API, providing a single call to create,
 // run, and remove an exec session. Returns exit code and error. Exit code is
 // not guaranteed to be set sanely if error is not nil.
-func (c *Container) Exec(config *ExecConfig, streams *define.AttachStreams, resize <-chan define.TerminalSize) (int, error) {
+func (c *Container) exec(config *ExecConfig, streams *define.AttachStreams, resizeChan <-chan resize.TerminalSize, isHealthcheck bool) (int, error) {
 	sessionID, err := c.ExecCreate(config)
 	if err != nil {
 		return -1, err
@@ -733,12 +774,15 @@ func (c *Container) Exec(config *ExecConfig, streams *define.AttachStreams, resi
 	// API there.
 	// TODO: Refactor so this is closed here, before we remove the exec
 	// session.
-	if resize != nil {
+	var size *resize.TerminalSize
+	if resizeChan != nil {
+		s := <-resizeChan
+		size = &s
 		go func() {
 			logrus.Debugf("Sending resize events to exec session %s", sessionID)
-			for resizeRequest := range resize {
+			for resizeRequest := range resizeChan {
 				if err := c.ExecResize(sessionID, resizeRequest); err != nil {
-					if errors.Cause(err) == define.ErrExecSessionStateInvalid {
+					if errors.Is(err, define.ErrExecSessionStateInvalid) {
 						// The exec session stopped
 						// before we could resize.
 						logrus.Infof("Missed resize on exec session %s, already stopped", sessionID)
@@ -751,29 +795,71 @@ func (c *Container) Exec(config *ExecConfig, streams *define.AttachStreams, resi
 		}()
 	}
 
-	if err := c.ExecStartAndAttach(sessionID, streams); err != nil {
+	if err := c.execStartAndAttach(sessionID, streams, size, isHealthcheck); err != nil {
 		return -1, err
 	}
 
-	session, err := c.ExecSession(sessionID)
+	session, err := c.execSessionNoCopy(sessionID)
 	if err != nil {
+		if errors.Is(err, define.ErrNoSuchExecSession) {
+			// TODO: If a proper Context is ever plumbed in here, we
+			// should use it.
+			// As things stand, though, it's not worth it - this
+			// should always terminate quickly since it's not
+			// streaming.
+			diedEvent, err := c.runtime.GetExecDiedEvent(context.Background(), c.ID(), sessionID)
+			if err != nil {
+				return -1, fmt.Errorf("retrieving exec session %s exit code: %w", sessionID, err)
+			}
+			return diedEvent.ContainerExitCode, nil
+		}
 		return -1, err
 	}
 	exitCode := session.ExitCode
 	if err := c.ExecRemove(sessionID, false); err != nil {
+		if errors.Is(err, define.ErrNoSuchExecSession) {
+			return exitCode, nil
+		}
 		return -1, err
 	}
 
 	return exitCode, nil
 }
 
-// cleanup an exec session after its done
-func (c *Container) cleanupExecBundle(sessionID string) error {
-	if err := os.RemoveAll(c.execBundlePath(sessionID)); err != nil && !os.IsNotExist(err) {
-		return err
+// cleanupExecBundle cleanups an exec session after its done
+// Please be careful when using this function since it might temporarily unlock
+// the container when os.RemoveAll($bundlePath) fails with ENOTEMPTY or EBUSY
+// errors.
+func (c *Container) cleanupExecBundle(sessionID string) (err error) {
+	path := c.execBundlePath(sessionID)
+	for attempts := 0; attempts < 50; attempts++ {
+		err = os.RemoveAll(path)
+		if err == nil || os.IsNotExist(err) {
+			return nil
+		}
+		if pathErr, ok := err.(*os.PathError); ok {
+			err = pathErr.Err
+			if errors.Is(err, unix.ENOTEMPTY) || errors.Is(err, unix.EBUSY) {
+				// give other processes a chance to use the container
+				if !c.batched {
+					if err := c.save(); err != nil {
+						return err
+					}
+					c.lock.Unlock()
+				}
+				time.Sleep(time.Millisecond * 100)
+				if !c.batched {
+					c.lock.Lock()
+					if err := c.syncContainer(); err != nil {
+						return err
+					}
+				}
+				continue
+			}
+		}
+		return
 	}
-
-	return nil
+	return
 }
 
 // the path to a containers exec session bundle
@@ -818,14 +904,14 @@ func (c *Container) createExecBundle(sessionID string) (retErr error) {
 	defer func() {
 		if retErr != nil {
 			if err := os.RemoveAll(bundlePath); err != nil {
-				logrus.Warnf("error removing exec bundle after creation caused another error: %v", err)
+				logrus.Warnf("Error removing exec bundle after creation caused another error: %v", err)
 			}
 		}
 	}()
 	if err := os.MkdirAll(c.execExitFileDir(sessionID), execDirPermission); err != nil {
 		// The directory is allowed to exist
 		if !os.IsExist(err) {
-			return errors.Wrapf(err, "error creating OCI runtime exit file path %s", c.execExitFileDir(sessionID))
+			return fmt.Errorf("creating OCI runtime exit file path %s: %w", c.execExitFileDir(sessionID), err)
 		}
 	}
 	return nil
@@ -838,11 +924,11 @@ func (c *Container) readExecExitCode(sessionID string) (int, error) {
 	chWait := make(chan error)
 	defer close(chWait)
 
-	_, err := WaitForFile(exitFile, chWait, time.Second*5)
+	_, err := util.WaitForFile(exitFile, chWait, time.Second*5)
 	if err != nil {
 		return -1, err
 	}
-	ec, err := ioutil.ReadFile(exitFile)
+	ec, err := os.ReadFile(exitFile)
 	if err != nil {
 		return -1, err
 	}
@@ -864,7 +950,7 @@ func (c *Container) getExecSessionPID(sessionID string) (int, error) {
 		return oldSession.PID, nil
 	}
 
-	return -1, errors.Wrapf(define.ErrNoSuchExecSession, "no exec session with ID %s found in container %s", sessionID, c.ID())
+	return -1, fmt.Errorf("no exec session with ID %s found in container %s: %w", sessionID, c.ID(), define.ErrNoSuchExecSession)
 }
 
 // getKnownExecSessions gets a list of all exec sessions we think are running,
@@ -904,7 +990,7 @@ func (c *Container) getActiveExecSessions() ([]string, error) {
 		alive, err := c.ociRuntime.ExecUpdateStatus(c, id)
 		if err != nil {
 			if lastErr != nil {
-				logrus.Errorf("Error checking container %s exec sessions: %v", c.ID(), lastErr)
+				logrus.Errorf("Checking container %s exec sessions: %v", c.ID(), lastErr)
 			}
 			lastErr = err
 			continue
@@ -919,7 +1005,7 @@ func (c *Container) getActiveExecSessions() ([]string, error) {
 				exitCode, err := c.readExecExitCode(session.ID())
 				if err != nil {
 					if lastErr != nil {
-						logrus.Errorf("Error checking container %s exec sessions: %v", c.ID(), lastErr)
+						logrus.Errorf("Checking container %s exec sessions: %v", c.ID(), lastErr)
 					}
 					lastErr = err
 				}
@@ -927,11 +1013,13 @@ func (c *Container) getActiveExecSessions() ([]string, error) {
 				session.PID = 0
 				session.State = define.ExecStateStopped
 
+				c.newExecDiedEvent(session.ID(), exitCode)
+
 				needSave = true
 			}
 			if err := c.cleanupExecBundle(id); err != nil {
 				if lastErr != nil {
-					logrus.Errorf("Error checking container %s exec sessions: %v", c.ID(), lastErr)
+					logrus.Errorf("Checking container %s exec sessions: %v", c.ID(), lastErr)
 				}
 				lastErr = err
 			}
@@ -942,7 +1030,7 @@ func (c *Container) getActiveExecSessions() ([]string, error) {
 	if needSave {
 		if err := c.save(); err != nil {
 			if lastErr != nil {
-				logrus.Errorf("Error reaping exec sessions for container %s: %v", c.ID(), lastErr)
+				logrus.Errorf("Reaping exec sessions for container %s: %v", c.ID(), lastErr)
 			}
 			lastErr = err
 		}
@@ -961,7 +1049,7 @@ func (c *Container) removeAllExecSessions() error {
 	for _, id := range knownSessions {
 		if err := c.ociRuntime.ExecStopContainer(c, id, c.StopTimeout()); err != nil {
 			if lastErr != nil {
-				logrus.Errorf("Error stopping container %s exec sessions: %v", c.ID(), lastErr)
+				logrus.Errorf("Stopping container %s exec sessions: %v", c.ID(), lastErr)
 			}
 			lastErr = err
 			continue
@@ -969,30 +1057,22 @@ func (c *Container) removeAllExecSessions() error {
 
 		if err := c.cleanupExecBundle(id); err != nil {
 			if lastErr != nil {
-				logrus.Errorf("Error stopping container %s exec sessions: %v", c.ID(), lastErr)
+				logrus.Errorf("Stopping container %s exec sessions: %v", c.ID(), lastErr)
 			}
 			lastErr = err
 		}
 	}
 	// Delete all exec sessions
 	if err := c.runtime.state.RemoveContainerExecSessions(c); err != nil {
-		if errors.Cause(err) != define.ErrCtrRemoved {
+		if !errors.Is(err, define.ErrCtrRemoved) {
 			if lastErr != nil {
-				logrus.Errorf("Error stopping container %s exec sessions: %v", c.ID(), lastErr)
+				logrus.Errorf("Stopping container %s exec sessions: %v", c.ID(), lastErr)
 			}
 			lastErr = err
 		}
 	}
 	c.state.ExecSessions = nil
 	c.state.LegacyExecSessions = nil
-	if err := c.save(); err != nil {
-		if errors.Cause(err) != define.ErrCtrRemoved {
-			if lastErr != nil {
-				logrus.Errorf("Error stopping container %s exec sessions: %v", c.ID(), lastErr)
-			}
-			lastErr = err
-		}
-	}
 
 	return lastErr
 }
@@ -1027,15 +1107,31 @@ func writeExecExitCode(c *Container, sessionID string, exitCode int) error {
 	// If we can't do this, no point in continuing, any attempt to save
 	// would write garbage to the DB.
 	if err := c.syncContainer(); err != nil {
-		if errors.Cause(err) == define.ErrNoSuchCtr || errors.Cause(err) == define.ErrCtrRemoved {
+		if errors.Is(err, define.ErrNoSuchCtr) || errors.Is(err, define.ErrCtrRemoved) {
 			// Container's entirely removed. We can't save status,
 			// but the container's entirely removed, so we don't
 			// need to. Exit without error.
 			return nil
 		}
-		return errors.Wrapf(err, "error syncing container %s state to remove exec session %s", c.ID(), sessionID)
+		return fmt.Errorf("syncing container %s state to remove exec session %s: %w", c.ID(), sessionID, err)
 	}
 
+	return justWriteExecExitCode(c, sessionID, exitCode)
+}
+
+func retrieveAndWriteExecExitCode(c *Container, sessionID string) error {
+	exitCode, err := c.readExecExitCode(sessionID)
+	if err != nil {
+		return err
+	}
+
+	return justWriteExecExitCode(c, sessionID, exitCode)
+}
+
+func justWriteExecExitCode(c *Container, sessionID string, exitCode int) error {
+	// Write an event first
+	c.newExecDiedEvent(sessionID, exitCode)
+
 	session, ok := c.state.ExecSessions[sessionID]
 	if !ok {
 		// Exec session already removed.
